"""
'pyenv' is recommended to use for testing with different Python versions
https://www.tecmint.com/pyenv-install-and-manage-multiple-python-versions-in-linux/

To get test coverage use 'coverage':
    $  coverage run -m stm32pio.tests.test -b
    $  coverage html
"""

import configparser
import inspect
import pathlib
import platform
import re
import shutil
import subprocess
import sys
import tempfile
import time
import unittest

import stm32pio.app
import stm32pio.lib
import stm32pio.settings


STM32PIO_MAIN_SCRIPT = inspect.getfile(stm32pio.app)  # absolute path to the main stm32pio script
# absolute path to the Python executable (no need to guess whether it's python or python3 and so on)
PYTHON_EXEC = sys.executable

# Test data
TEST_PROJECT_PATH = pathlib.Path('stm32pio-test-project').resolve()
if not TEST_PROJECT_PATH.is_dir() or not TEST_PROJECT_PATH.joinpath('stm32pio-test-project.ioc').is_file():
    raise FileNotFoundError("No test project is present")
# Make sure you have F0 framework installed (try to run code generation from STM32CubeMX manually at least once before
# proceeding)
TEST_PROJECT_BOARD = 'nucleo_f031k6'

# Instantiate a temporary folder on every fixture run. It is used across all tests and is deleted on shutdown
temp_dir = tempfile.TemporaryDirectory()
FIXTURE_PATH = pathlib.Path(temp_dir.name).joinpath(TEST_PROJECT_PATH.name)
print(f"Temp test fixture path: {FIXTURE_PATH}")


class CustomTestCase(unittest.TestCase):
    """
    These pre- and post-tasks are common for all test cases
    """

    def setUp(self):
        """
        Copy the test project from the repo to our temp directory. WARNING: make sure the test project folder is clean
        (i.e. contains only an .ioc file) before running the test
        """
        shutil.rmtree(FIXTURE_PATH, ignore_errors=True)
        shutil.copytree(TEST_PROJECT_PATH, FIXTURE_PATH)

    def tearDown(self):
        """
        Clean the temp directory
        """
        shutil.rmtree(FIXTURE_PATH, ignore_errors=True)


class TestUnit(CustomTestCase):
    """
    Test the single method. As we at some point decided to use a class instead of the set of scattered functions we need
    to do some preparations for almost every test (e.g. instantiate the class, create the PlatformIO project, etc.),
    though
    """

    def test_generate_code(self):
        """
        Check whether files and folders have been created by STM32CubeMX
        """
        project = stm32pio.lib.Stm32pio(FIXTURE_PATH, parameters={'board': TEST_PROJECT_BOARD},
                                        save_on_destruction=False)
        project.generate_code()

        # Assuming that the presence of these files indicating a success
        files_should_be_present = ['Src/main.c', 'Inc/main.h']
        for file in files_should_be_present:
            with self.subTest(file_should_be_present=file, msg=f"{file} hasn't been created"):
                self.assertEqual(FIXTURE_PATH.joinpath(file).is_file(), True)

    def test_pio_init(self):
        """
        Consider that existence of 'platformio.ini' file showing a successful PlatformIO project initialization. The
        last one has another traces those can be checked too but we are interested only in a 'platformio.ini' anyway.
        Also, check that it is a correct configparser file and is not empty
        """
        project = stm32pio.lib.Stm32pio(FIXTURE_PATH, parameters={'board': TEST_PROJECT_BOARD},
                                        save_on_destruction=False)
        result = project.pio_init()

        self.assertEqual(result, 0, msg="Non-zero return code")
        self.assertTrue(FIXTURE_PATH.joinpath('platformio.ini').is_file(), msg="platformio.ini is not there")

        platformio_ini = configparser.ConfigParser(interpolation=None)
        self.assertGreater(len(platformio_ini.read(str(FIXTURE_PATH.joinpath('platformio.ini')))), 0,
                           msg='platformio.ini is empty')

    def test_patch(self):
        """
        Check that new parameters were added, modified were updated and existing parameters didn't gone. Also, check for
        unnecessary folders deletion
        """

        project = stm32pio.lib.Stm32pio(FIXTURE_PATH, save_on_destruction=False)

        # We do not create a real project here so we don't depend on other possible issues
        test_content = inspect.cleandoc('''
            ; This is a test config .ini file
            ; with a comment. It emulates a real
            ; platformio.ini file

            [platformio]
            include_dir = this s;789hould be replaced
            ; there should appear a new parameter
            test_key3 = this should be preserved

            [test_section]
            test_key1 = test_value1
            test_key2 = 123
        ''') + '\n'
        FIXTURE_PATH.joinpath('platformio.ini').write_text(test_content)
        FIXTURE_PATH.joinpath('include').mkdir()

        project.patch()

        self.assertFalse(FIXTURE_PATH.joinpath('include').is_dir(), msg="'include' has not been deleted")

        original_test_config = configparser.ConfigParser(interpolation=None)
        original_test_config.read_string(test_content)

        patched_config = configparser.ConfigParser(interpolation=None)
        patch_config = configparser.ConfigParser(interpolation=None)
        patch_config.read_string(project.config.get('project', 'platformio_ini_patch_content'))

        self.assertGreater(len(patched_config.read(FIXTURE_PATH.joinpath('platformio.ini'))), 0)

        for patch_section in patch_config.sections():
            self.assertTrue(patched_config.has_section(patch_section), msg=f"{patch_section} is missing")
            for patch_key, patch_value in patch_config.items(patch_section):
                self.assertEqual(patched_config.get(patch_section, patch_key, fallback=None), patch_value,
                                 msg=f"{patch_section}: {patch_key}={patch_value} is missing or incorrect in the "
                                     "patched config")

        for original_section in original_test_config.sections():
            self.assertTrue(patched_config.has_section(original_section),
                            msg=f"{original_section} from the original config is missing")
            for original_key, original_value in original_test_config.items(original_section):
                # We've already checked patch parameters so skip them
                if not patch_config.has_option(original_section, original_key):
                    self.assertEqual(patched_config.get(original_section, original_key), original_value,
                                     msg=f"{original_section}: {original_key}={original_value} is corrupted")

    def test_build_should_handle_error(self):
        """
        Build an empty project so PlatformIO should return the error
        """
        project = stm32pio.lib.Stm32pio(FIXTURE_PATH, parameters={'board': TEST_PROJECT_BOARD},
                                        save_on_destruction=False)
        project.pio_init()

        with self.assertLogs(level='ERROR') as logs:
            self.assertNotEqual(project.pio_build(), 0, msg="Build error was not indicated")
            self.assertTrue(next((True for item in logs.output if "PlatformIO build error" in item), False),
                            msg="Error message does not match")

    def test_run_editor(self):
        """
        Call the editors
        """
        project = stm32pio.lib.Stm32pio(FIXTURE_PATH, save_on_destruction=False)

        editors = {
            'atom': {
                'Windows': 'atom.exe',
                'Darwin': 'Atom',
                'Linux': 'atom'
            },
            'code': {
                'Windows': 'Code.exe',
                'Darwin': 'Visual Studio Code',
                'Linux': 'code'
            },
            'subl': {
                'Windows': 'sublime_text.exe',
                'Darwin': 'Sublime',
                'Linux': 'sublime'
            }
        }

        for command, name in editors.items():
            # Look for the command presence in the system so we test only installed editors
            if platform.system() == 'Windows':
                command_str = f"where {command} /q"
            else:
                command_str = f"command -v {command}"
            editor_exists = False
            if subprocess.run(command_str, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).returncode == 0:
                editor_exists = True

            if editor_exists:
                with self.subTest(command=command, name=name[platform.system()]):
                    project.start_editor(command)

                    time.sleep(1)  # wait a little bit for app to start

                    command_arr = ['ps', '-A']
                    if platform.system() == 'Windows':
                        command_arr = ['wmic', 'process', 'get', 'description']
                    # "encoding='utf-8'" is for "a bytes-like object is required, not 'str'" in "assertIn"
                    result = subprocess.run(command_arr, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                            encoding='utf-8')
                    # Or, for Python 3.7 and above:
                    # result = subprocess.run(command_arr, capture_output=True, encoding='utf-8')
                    self.assertIn(name[platform.system()], result.stdout)

    def test_init_path_not_found_should_raise(self):
        """
        Pass non-existing path and expect the error
        """
        path_does_not_exist_name = 'does_not_exist'

        path_does_not_exist = FIXTURE_PATH.joinpath(path_does_not_exist_name)
        with self.assertRaisesRegex(FileNotFoundError, path_does_not_exist_name,
                                    msg="FileNotFoundError was not raised or doesn't contain a description"):
            stm32pio.lib.Stm32pio(path_does_not_exist, save_on_destruction=False)

    def test_save_config(self):
        """
        Explicitly save the config to file and look did that actually happen and whether all the information was
        preserved
        """
        # 'board' is non-default, 'project'-section parameter
        project = stm32pio.lib.Stm32pio(FIXTURE_PATH, parameters={'board': TEST_PROJECT_BOARD},
                                        save_on_destruction=False)

        project.config.save()

        self.assertTrue(FIXTURE_PATH.joinpath(stm32pio.settings.config_file_name).is_file(),
                        msg=f"{stm32pio.settings.config_file_name} file hasn't been created")

        config = configparser.ConfigParser(interpolation=None)
        config.read(str(FIXTURE_PATH.joinpath(stm32pio.settings.config_file_name)))
        for section, parameters in stm32pio.settings.config_default.items():
            for option, value in parameters.items():
                with self.subTest(section=section, option=option, msg="Section/key is not found in saved config file"):
                    self.assertNotEqual(config.get(section, option, fallback="Not found"), "Not found")
        self.assertEqual(config.get('project', 'board', fallback="Not found"), TEST_PROJECT_BOARD,
                         msg="'board' has not been set")


class TestIntegration(CustomTestCase):
    """
    Sequence of methods that should work seamlessly
    """

    def test_config_priorities(self):
        """
        Test the compliance with priorities when reading the parameters
        """

        # Sample user's custom patch value
        config_parameter_user_value = inspect.cleandoc('''
            [test_section]
            key1 = value1
            key2 = 789
        ''')
        cli_parameter_user_value = 'nucleo_f429zi'

        # Create test config
        config = configparser.ConfigParser(interpolation=None)
        config.read_dict({
            'project': {
                'platformio_ini_patch_content': config_parameter_user_value,
                'board': TEST_PROJECT_BOARD
            }
        })
        # ... save it
        with FIXTURE_PATH.joinpath(stm32pio.settings.config_file_name).open(mode='w') as config_file:
            config.write(config_file)

        # On project creation we should interpret the CLI-provided values as superseding to the saved ones and
        # saved ones as superseding to the default ones
        project = stm32pio.lib.Stm32pio(FIXTURE_PATH, parameters={'board': cli_parameter_user_value},
                                        save_on_destruction=False)
        project.pio_init()
        project.patch()

        # Actually, we can parse platformio.ini via configparser but this is simpler
        after_patch_content = FIXTURE_PATH.joinpath('platformio.ini').read_text()
        self.assertIn(config_parameter_user_value, after_patch_content,
                      msg="User config parameter has not been prioritized over the default one")
        self.assertIn(cli_parameter_user_value, after_patch_content,
                      msg="User CLI parameter has not been prioritized over the saved one")

    def test_build(self):
        """
        Initialize a new project and try to build it
        """
        project = stm32pio.lib.Stm32pio(FIXTURE_PATH, parameters={'board': TEST_PROJECT_BOARD},
                                        save_on_destruction=False)
        project.generate_code()
        project.pio_init()
        project.patch()

        result = project.pio_build()

        self.assertEqual(result, 0, msg="Build failed")

    def test_regenerate_code(self):
        """
        Simulate a new project creation, its changing and CubeMX code re-generation (for example, after adding new
        hardware features and some new files by a user)
        """
        project = stm32pio.lib.Stm32pio(FIXTURE_PATH, parameters={'board': TEST_PROJECT_BOARD},
                                        save_on_destruction=False)

        # Generate a new project ...
        project.generate_code()
        project.pio_init()
        project.patch()

        # ... change it:
        test_file_1 = FIXTURE_PATH.joinpath('Src', 'main.c')
        test_content_1 = "*** TEST STRING 1 ***\n"
        test_file_2 = FIXTURE_PATH.joinpath('Inc', 'my_header.h')
        test_content_2 = "*** TEST STRING 2 ***\n"
        #   - add some sample string inside CubeMX' /* BEGIN - END */ block
        main_c_content = test_file_1.read_text()
        pos = main_c_content.index("while (1)")
        main_c_new_content = main_c_content[:pos] + test_content_1 + main_c_content[pos:]
        test_file_1.write_text(main_c_new_content)
        #  - add new file inside the project
        test_file_2.write_text(test_content_2)

        # Re-generate CubeMX project
        project.generate_code()

        # Check if added information is preserved
        main_c_after_regenerate_content = test_file_1.read_text()
        my_header_h_after_regenerate_content = test_file_2.read_text()
        self.assertIn(test_content_1, main_c_after_regenerate_content,
                      msg=f"User content hasn't been preserved after regeneration in {test_file_1}")
        self.assertIn(test_content_2, my_header_h_after_regenerate_content,
                      msg=f"User content hasn't been preserved after regeneration in {test_file_2}")

    def test_get_state(self):
        """
        Go through the sequence of states emulating the real-life project lifecycle
        """

        project = stm32pio.lib.Stm32pio(FIXTURE_PATH, parameters={'board': TEST_PROJECT_BOARD},
                                        save_on_destruction=False)
        self.assertEqual(project.state, stm32pio.lib.ProjectState.UNDEFINED)

        project.config.save()
        self.assertEqual(project.state, stm32pio.lib.ProjectState.INITIALIZED)

        project.generate_code()
        self.assertEqual(project.state, stm32pio.lib.ProjectState.GENERATED)

        project.pio_init()
        self.assertEqual(project.state, stm32pio.lib.ProjectState.PIO_INITIALIZED)

        project.patch()
        self.assertEqual(project.state, stm32pio.lib.ProjectState.PATCHED)

        project.pio_build()
        self.assertEqual(project.state, stm32pio.lib.ProjectState.BUILT)

        project.clean()
        self.assertEqual(project.state, stm32pio.lib.ProjectState.UNDEFINED)


class TestCLI(CustomTestCase):
    """
    Some tests to mimic the behavior of end-user tasks (CLI commands such as 'new', 'clean', etc.). Run main function
    passing the arguments to it but sometimes even run as subprocess (to capture actual STDOUT/STDERR output)
    """

    def test_clean(self):
        # Create files and folders
        file_should_be_deleted = FIXTURE_PATH.joinpath('file.should.be.deleted')
        dir_should_be_deleted = FIXTURE_PATH.joinpath('dir.should.be.deleted')
        file_should_be_deleted.touch(exist_ok=False)
        dir_should_be_deleted.mkdir(exist_ok=False)

        # Clean
        return_code = stm32pio.app.main(sys_argv=['clean', '-d', str(FIXTURE_PATH)])
        self.assertEqual(return_code, 0, msg="Non-zero return code")

        # Look for remaining items
        self.assertFalse(file_should_be_deleted.is_file(), msg=f"{file_should_be_deleted} is still there")
        self.assertFalse(dir_should_be_deleted.is_dir(), msg=f"{dir_should_be_deleted} is still there")

        # And .ioc file should be preserved
        self.assertTrue(FIXTURE_PATH.joinpath(f"{FIXTURE_PATH.name}.ioc").is_file(), msg="Missing .ioc file")

    def test_new(self):
        """
        Successful build is the best indicator that all went right so we use '--with-build' option
        """

        return_code = stm32pio.app.main(sys_argv=['new', '-d', str(FIXTURE_PATH), '-b', TEST_PROJECT_BOARD,
                                                  '--with-build'])
        self.assertEqual(return_code, 0, msg="Non-zero return code")

        # .ioc file should be preserved
        self.assertTrue(FIXTURE_PATH.joinpath(f"{FIXTURE_PATH.name}.ioc").is_file(), msg="Missing .ioc file")

    def test_generate(self):
        return_code = stm32pio.app.main(sys_argv=['generate', '-d', str(FIXTURE_PATH)])
        self.assertEqual(return_code, 0, msg="Non-zero return code")

        inc_dir = 'Inc'
        src_dir = 'Src'

        self.assertTrue(FIXTURE_PATH.joinpath(inc_dir).is_dir(), msg=f"Missing '{inc_dir}'")
        self.assertTrue(FIXTURE_PATH.joinpath(src_dir).is_dir(), msg=f"Missing '{src_dir}'")
        self.assertFalse(len(list(FIXTURE_PATH.joinpath(inc_dir).iterdir())) == 0, msg=f"'{inc_dir}' is empty")
        self.assertFalse(len(list(FIXTURE_PATH.joinpath(src_dir).iterdir())) == 0, msg=f"'{src_dir}' is empty")

        # .ioc file should be preserved
        self.assertTrue(FIXTURE_PATH.joinpath(f"{FIXTURE_PATH.name}.ioc").is_file(), msg="Missing .ioc file")

    def test_incorrect_path_should_log_error(self):
        """
        We should see an error log message and non-zero return code
        """

        path_not_exist = pathlib.Path('path/does/not/exist')

        with self.assertLogs(level='ERROR') as logs:
            return_code = stm32pio.app.main(sys_argv=['init', '-d', str(path_not_exist)])
            self.assertNotEqual(return_code, 0, msg='Return code should be non-zero')
            self.assertTrue(next((True for item in logs.output if str(path_not_exist) in item), False),
                            msg="'ERROR' logging message hasn't been printed")

    def test_no_ioc_file_should_log_error(self):
        """
        We should see an error log message and non-zero return code
        """

        dir_with_no_ioc_file = FIXTURE_PATH.joinpath('dir.with.no.ioc.file')
        dir_with_no_ioc_file.mkdir(exist_ok=False)

        with self.assertLogs(level='ERROR') as logs:
            return_code = stm32pio.app.main(sys_argv=['init', '-d', str(dir_with_no_ioc_file)])
            self.assertNotEqual(return_code, 0, msg='Return code should be non-zero')
            self.assertTrue(next((True for item in logs.output if "CubeMX project .ioc file" in item), False),
                            msg="'ERROR' logging message hasn't been printed")

    def test_verbose(self):
        """
        Run as subprocess to capture the full output. Check for both 'DEBUG' logging messages and STM32CubeMX CLI
        output. Verbose logs format should match such a regex:

            ^(?=(DEBUG|INFO|WARNING|ERROR|CRITICAL) {0,4})(?=.{8} (?=(pio_build|pio_init|...) {0,26})(?=.{26} [^ ]))

        """

        project = stm32pio.lib.Stm32pio(FIXTURE_PATH, save_on_destruction=False)
        methods = [method[0] for method in inspect.getmembers(project, predicate=inspect.ismethod)]
        methods.append('main')

        result = subprocess.run([PYTHON_EXEC, STM32PIO_MAIN_SCRIPT, '-v', 'generate', '-d', str(FIXTURE_PATH)],
                                encoding='utf-8', stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        self.assertEqual(result.returncode, 0, msg="Non-zero return code")
        # Somehow stderr and not stdout contains the actual output but we check both
        self.assertTrue('DEBUG' in result.stderr or 'DEBUG' in result.stdout,
                        msg="Verbose logging output hasn't been enabled on stderr")
        regex = re.compile("^(?=(DEBUG) {0,4})(?=.{8} (?=(" + '|'.join(methods) + ") {0,26})(?=.{26} [^ ]))",
                           flags=re.MULTILINE)
        self.assertGreaterEqual(len(re.findall(regex, result.stderr)), 1,
                                msg="Logs messages doesn't match the format")

        self.assertIn('Starting STM32CubeMX', result.stdout, msg="STM32CubeMX didn't print its logs")

    def test_non_verbose(self):
        """
        Run as subprocess to capture the full output. We should not see any 'DEBUG' logging messages or STM32CubeMX CLI
        output. Logs format should match such a regex:

            ^(?=(INFO) {0,4})(?=.{8} ((?!( |pio_build|pio_init|...))))

        """

        project = stm32pio.lib.Stm32pio(FIXTURE_PATH, save_on_destruction=False)
        methods = [method[0] for method in inspect.getmembers(project, predicate=inspect.ismethod)]
        methods.append('main')

        result = subprocess.run([PYTHON_EXEC, STM32PIO_MAIN_SCRIPT, 'generate', '-d', str(FIXTURE_PATH)],
                                encoding='utf-8', stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        self.assertEqual(result.returncode, 0, msg="Non-zero return code")
        self.assertNotIn('DEBUG', result.stderr, msg="Verbose logging output has been enabled on stderr")
        self.assertNotIn('DEBUG', result.stdout, msg="Verbose logging output has been enabled on stdout")

        regex = re.compile("^(?=(INFO) {0,4})(?=.{8} ((?!( |" + '|'.join(methods) + "))))", flags=re.MULTILINE)
        self.assertGreaterEqual(len(re.findall(regex, result.stderr)), 1,
                                msg="Logs messages doesn't match the format")

        self.assertNotIn('Starting STM32CubeMX', result.stdout, msg="STM32CubeMX printed its logs")

    def test_init(self):
        """
        Check for config creation and parameters presence
        """

        result = subprocess.run([PYTHON_EXEC, STM32PIO_MAIN_SCRIPT, 'init', '-d', str(FIXTURE_PATH),
                                 '-b', TEST_PROJECT_BOARD])
        self.assertEqual(result.returncode, 0, msg="Non-zero return code")

        self.assertTrue(FIXTURE_PATH.joinpath(stm32pio.settings.config_file_name).is_file(),
                        msg=f"{stm32pio.settings.config_file_name} file hasn't been created")

        config = configparser.ConfigParser(interpolation=None)
        config.read(str(FIXTURE_PATH.joinpath(stm32pio.settings.config_file_name)))
        for section, parameters in stm32pio.settings.config_default.items():
            for option, value in parameters.items():
                with self.subTest(section=section, option=option, msg="Section/key is not found in saved config file"):
                    self.assertIsNotNone(config.get(section, option, fallback=None))
        self.assertEqual(config.get('project', 'board', fallback="Not found"), TEST_PROJECT_BOARD,
                         msg="'board' has not been set")


if __name__ == '__main__':
    unittest.main()



"""
Main library
"""

import collections
import configparser
import enum
import logging
import pathlib
import shutil
import string
import subprocess
import tempfile
import weakref

import stm32pio.settings

# Child logger
logger = logging.getLogger('stm32pio.util')


@enum.unique
class ProjectState(enum.IntEnum):
    """
    Codes indicating a project state at the moment. Should be the sequence of incrementing integers to be suited for
    state determining algorithm

    Hint: Files/folders to be present on every project state:
        UNDEFINED: use this state to indicate none of the states below. Also, when we do not have any .ioc file the
                   Stm32pio class cannot be instantiated (constructor raises an exception)
        INITIALIZED: ['project.ioc', 'stm32pio.ini']
        GENERATED: ['Inc', 'Src', 'project.ioc', 'stm32pio.ini']
        PIO_INITIALIZED (on case-sensitive FS): ['test', 'include', 'Inc', 'platformio.ini', '.gitignore', 'Src', 'lib',
                                                 'project.ioc', '.travis.yml', 'src']
        PIO_INI_PATCHED: ['test', 'Inc', 'platformio.ini', '.gitignore', 'Src', 'lib', 'project.ioc', '.travis.yml']
        BUILT: same as above + '.pio' folder with build artifacts (such as .pio/build/nucleo_f031k6/firmware.bin,
                                                                           .pio/build/nucleo_f031k6/firmware.elf)
    """
    UNDEFINED = enum.auto()
    INITIALIZED = enum.auto()
    GENERATED = enum.auto()
    PIO_INITIALIZED = enum.auto()
    PATCHED = enum.auto()
    BUILT = enum.auto()


class Config(configparser.ConfigParser):
    def __init__(self, location: pathlib.Path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._location = location

    def save(self):
        """
        Tries to save the config to the file and gently log if any error occurs
        """
        try:
            with self._location.joinpath(stm32pio.settings.config_file_name).open(mode='w') as config_file:
                self.write(config_file)
            logger.debug("stm32pio.ini config file has been saved")
            return 0
        except Exception as e:
            logger.warning(f"cannot save the config: {e}", exc_info=logger.getEffectiveLevel() <= logging.DEBUG)
            return -1


class Stm32pio:
    """
    Main class.

    Represents a single project, encapsulating file system path to the project (main mandatory identifier) and some
    parameters in a configparser .ini file. As stm32pio can be installed via pip and has no global config we also
    storing global parameters (such as Java or STM32CubeMX invoking commands) in this config .ini file so the user can
    specify settings on a per-project base. The config can be saved in a non-disturbing way automatically on the
    instance destruction (e.g. by garbage collecting it) (use save_on_destruction=True flag), otherwise a user should
    explicitly save the config if he wants to (using config.save() method).

    The typical life cycle consists of project creation, passing mandatory 'dirty_path' argument. If also 'parameters'
    dictionary is specified also these settings are processed (white-list approach is used so we set only those
    parameters that are listed in the constructor code) (currently only 'board' parameter is included). Then it is
    possible to perform API operations. WARNING. Please be careful with the 'clean' method as it deletes all the content
    of the project directory except the main .ioc file.

    Args:
        dirty_path (str): path to the project
        parameters (dict): additional parameters to set on initialization stage
        save_on_destruction (bool): register or not the finalizer that saves the config to file
    """

    def __init__(self, dirty_path: str, parameters: dict = None, save_on_destruction: bool = True):
        if parameters is None:
            parameters = {}

        # The path is a unique identifier of the project so it would be great to remake Stm32pio class as a subclass of
        # pathlib.Path and then reference it like self and not self.project_path. It is more consistent also, as now
        # project_path is perceived like any other config parameter that somehow is appeared to exist outside of a
        # config instance but then it will be a core identifier, a truly 'self' value.
        #
        # But currently pathlib.Path is not intended to be subclassable by-design, unfortunately. See
        # https://bugs.python.org/issue24132
        self.project_path = self._resolve_project_path(dirty_path)
        self.config = self._load_config_file()

        ioc_file = self._find_ioc_file()
        self.config.set('project', 'ioc_file', str(ioc_file))

        cubemx_script_template = string.Template(self.config.get('project', 'cubemx_script_content'))
        cubemx_script_content = cubemx_script_template.substitute(project_path=self.project_path,
            cubemx_ioc_full_filename=self.config.get('project', 'ioc_file'))
        self.config.set('project', 'cubemx_script_content', cubemx_script_content)

        board = ''
        if 'board' in parameters and parameters['board'] is not None:
            try:
                board = self._resolve_board(parameters['board'])
            except Exception as e:
                logger.warning(e)
            self.config.set('project', 'board', board)
        elif self.config.get('project', 'board', fallback=None) is None:
            self.config.set('project', 'board', board)

        if save_on_destruction:
            self._finalizer = weakref.finalize(self, self.config.save)


    @property
    def state(self) -> ProjectState:
        """
        Property returning the current state of the project. Calculated at every request.
        """

        logger.debug("calculating the project state...")
        logger.debug(f"project content: {[item.name for item in self.project_path.iterdir()]}")

        try:
            platformio_ini_is_patched = self.platformio_ini_is_patched()
        except:
            platformio_ini_is_patched = False

        states_conditions = collections.OrderedDict()
        # Fill the ordered dictionary with conditions results
        states_conditions[ProjectState.UNDEFINED] = [True]
        states_conditions[ProjectState.INITIALIZED] = [
            self.project_path.joinpath(stm32pio.settings.config_file_name).is_file()]
        states_conditions[ProjectState.GENERATED] = [self.project_path.joinpath('Inc').is_dir() and
                                                     len(list(self.project_path.joinpath('Inc').iterdir())) > 0,
                                                     self.project_path.joinpath('Src').is_dir() and
                                                     len(list(self.project_path.joinpath('Src').iterdir())) > 0]
        states_conditions[ProjectState.PIO_INITIALIZED] = [
            self.project_path.joinpath('platformio.ini').is_file() and
            len(self.project_path.joinpath('platformio.ini').read_text()) > 0]
        states_conditions[ProjectState.PATCHED] = [
            platformio_ini_is_patched, not self.project_path.joinpath('include').is_dir()]
        states_conditions[ProjectState.BUILT] = [
            self.project_path.joinpath('.pio').is_dir() and
            any([item.is_file() for item in self.project_path.joinpath('.pio').rglob('*firmware*')])]

        # Use (1,0) instead of (True,False) because on debug printing it looks better
        conditions_results = []
        for state, conditions in states_conditions.items():
            conditions_results.append(1 if all(condition is True for condition in conditions) else 0)

        # Put away unnecessary processing as the string still will be formed even if the logging level doesn't allow
        # propagation of this message
        if logger.getEffectiveLevel() <= logging.DEBUG:
            states_info_str = '\n'.join(
                f"{state.name:20}{conditions_results[state.value - 1]}" for state in ProjectState)
            logger.debug(f"determined states:\n{states_info_str}")

        # Search for a consecutive raw of 1's and find the last of them. For example, if the array is
        #   [1,1,0,1,0,0]
        #      ^
        last_true_index = 0  # UNDEFINED is always True, use as a start value
        for index, value in enumerate(conditions_results):
            if value == 1:
                last_true_index = index
            else:
                break

        # Fall back to the UNDEFINED state if we have breaks in conditions results array. For example, in [1,1,0,1,0,0]
        # we still return UNDEFINED as it doesn't look like a correct combination of files actually
        project_state = ProjectState.UNDEFINED
        if 1 not in conditions_results[last_true_index + 1:]:
            project_state = ProjectState(last_true_index + 1)

        return project_state


    def _find_ioc_file(self) -> pathlib.Path:
        """
        Find and return an .ioc file. If there are more than one, return first. If no .ioc file is present raise
        FileNotFoundError exception
        """

        ioc_file = self.config.get('project', 'ioc_file', fallback=None)
        if ioc_file:
            return pathlib.Path(ioc_file).resolve()
        else:
            logger.debug("searching for any .ioc file...")
            candidates = list(self.project_path.glob('*.ioc'))
            if len(candidates) == 0:  # good candidate for the new Python 3.8 assignment expressions feature :)
                raise FileNotFoundError("not found: CubeMX project .ioc file")
            elif len(candidates) == 1:
                logger.debug(f"{candidates[0].name} is selected")
                return candidates[0]
            else:
                logger.warning(f"there are multiple .ioc files, {candidates[0].name} is selected")
                return candidates[0]


    def _load_config_file(self) -> Config:
        """
        Prepare configparser config for the project. First, read the default config and then mask these values with user
        ones
        """

        logger.debug(f"searching for {stm32pio.settings.config_file_name}...")
        stm32pio_ini = self.project_path.joinpath(stm32pio.settings.config_file_name)

        config = Config(self.project_path, interpolation=None)

        # Fill with default values
        config.read_dict(stm32pio.settings.config_default)
        # Then override by user values (if exist)
        config.read(str(stm32pio_ini))

        # Put away unnecessary processing as the string still will be formed even if the logging level doesn't allow
        # propagation of this message
        if logger.getEffectiveLevel() <= logging.DEBUG:
            debug_str = 'resolved config:'
            for section in config.sections():
                debug_str += f"\n========== {section} ==========\n"
                for value in config.items(section):
                    debug_str += f"{value}\n"
            logger.debug(debug_str)

        return config


    @staticmethod
    def _resolve_project_path(dirty_path: str) -> pathlib.Path:
        """
        Handle 'path/to/proj' and 'path/to/proj/', '.' (current directory) and other cases

        Args:
            dirty_path (str): some directory in the filesystem
        """
        resolved_path = pathlib.Path(dirty_path).expanduser().resolve()
        if not resolved_path.exists():
            raise FileNotFoundError(f"not found: {resolved_path}")
        else:
            return resolved_path


    def _resolve_board(self, board: str) -> str:
        """
        Check if given board is a correct board name in the PlatformIO database

        Returns:
            same board that has been given, raise an exception otherwise
        """

        logger.debug("searching for PlatformIO board...")
        result = subprocess.run([self.config.get('app', 'platformio_cmd'), 'boards'], encoding='utf-8',
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # Or, for Python 3.7 and above:
        # result = subprocess.run(['platformio', 'boards'], encoding='utf-8', capture_output=True)
        if result.returncode == 0:
            if board not in result.stdout.split():
                raise Exception("wrong PlatformIO STM32 board. Run 'platformio boards' for possible names")
            else:
                logger.debug(f"PlatformIO board {board} was found")
                return board
        else:
            raise Exception("failed to search for PlatformIO boards")


    def generate_code(self) -> None:
        """
        Call STM32CubeMX app as a 'java -jar' file to generate the code from the .ioc file. Pass commands to the
        STM32CubeMX in a temp file
        """

        # Use mkstemp() instead of higher-level API for compatibility with Windows (see tempfile docs for more details)
        cubemx_script_file, cubemx_script_name = tempfile.mkstemp()

        # buffering=0 leads to the immediate flushing on writing
        with open(cubemx_script_file, mode='w+b', buffering=0) as cubemx_script:
            cubemx_script.write(self.config.get('project', 'cubemx_script_content').encode())  # encode since mode='w+b'

            logger.info("starting to generate a code from the CubeMX .ioc file...")
            command_arr = [self.config.get('app', 'java_cmd'), '-jar', self.config.get('app', 'cubemx_cmd'), '-q',
                           cubemx_script_name, '-s']  # -q: read commands from file, -s: silent performance
            if logger.getEffectiveLevel() <= logging.DEBUG:
                result = subprocess.run(command_arr)
            else:
                result = subprocess.run(command_arr, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                # Or, for Python 3.7 and above:
                # result = subprocess.run(command_arr, capture_output=True)
            if result.returncode == 0:
                logger.info("successful code generation")
            else:
                logger.error(f"code generation error (CubeMX return code is {result.returncode}).\n"
                             "Enable a verbose output or try to generate a code from the CubeMX itself.")
                raise Exception("code generation error")

        pathlib.Path(cubemx_script_name).unlink()


    def pio_init(self) -> int:
        """
        Call PlatformIO CLI to initialize a new project. It uses parameters (path, board) collected before so the
        confirmation of the data presence is on a user
        """

        logger.info("starting PlatformIO project initialization...")

        platformio_ini_file = self.project_path.joinpath('platformio.ini')
        if platformio_ini_file.is_file() and len(platformio_ini_file.read_text()) > 0:
            logger.warning("'platformio.ini' file is already existing")

        # TODO: move out to config a 'framework' option and to settings a 'platformio.ini' file name
        command_arr = [self.config.get('app', 'platformio_cmd'), 'init', '-d', str(self.project_path), '-b',
                       self.config.get('project', 'board'), '-O', 'framework=stm32cube']
        if logger.getEffectiveLevel() > logging.DEBUG:
            command_arr.append('--silent')

        result = subprocess.run(command_arr, encoding='utf-8', stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        error_msg = "PlatformIO project initialization error"
        if result.returncode == 0:
            # PlatformIO returns 0 even on some errors (e.g. no '--board' argument)
            for output in [result.stdout, result.stderr]:
                if 'ERROR' in output.upper():
                    logger.error(output)
                    raise Exception(error_msg)
            logger.info("successful PlatformIO project initialization")
            return result.returncode
        else:
            raise Exception(error_msg)


    def platformio_ini_is_patched(self) -> bool:
        """
        Check whether 'platformio.ini' config file is patched or not. It doesn't check for unnecessary folders deletion
        """

        platformio_ini = configparser.ConfigParser(interpolation=None)
        try:
            if len(platformio_ini.read(self.project_path.joinpath('platformio.ini'))) == 0:
                raise FileNotFoundError("not found: 'platformio.ini' file")
        except FileNotFoundError as e:
            raise e
        except Exception as e:
            raise Exception("'platformio.ini' file is incorrect") from e

        patch_config = configparser.ConfigParser(interpolation=None)
        try:
            patch_config.read_string(self.config.get('project', 'platformio_ini_patch_content'))
        except Exception as e:
            raise Exception("Desired patch content is invalid (should satisfy INI-format requirements)") from e

        for patch_section in patch_config.sections():
            if platformio_ini.has_section(patch_section):
                for patch_key, patch_value in patch_config.items(patch_section):
                    if platformio_ini.get(patch_section, patch_key, fallback=None) != patch_value:
                        return False
            else:
                return False
        return True


    def patch(self) -> None:
        """
        Patch platformio.ini file by a user's patch. By default, it sets the created earlier (by CubeMX 'Src' and 'Inc')
        folders as sources. configparser doesn't preserve any comments unfortunately so keep in mid that all of them
        will be lost at this stage. Also, the order may be violated. In the end, remove old empty folders
        """

        logger.debug("patching 'platformio.ini' file...")

        if self.platformio_ini_is_patched():
            logger.info("'platformio.ini' has been already patched")
        else:
            # Existing .ini file
            platformio_ini_config = configparser.ConfigParser(interpolation=None)
            platformio_ini_config.read(self.project_path.joinpath('platformio.ini'))

            # Our patch has the config format too
            patch_config = configparser.ConfigParser(interpolation=None)
            patch_config.read_string(self.config.get('project', 'platformio_ini_patch_content'))

            # Merge 2 configs
            for patch_section in patch_config.sections():
                if not platformio_ini_config.has_section(patch_section):
                    platformio_ini_config.add_section(patch_section)
                for patch_key, patch_value in patch_config.items(patch_section):
                    platformio_ini_config.set(patch_section, patch_key, patch_value)

            # Save, overwriting the original file
            with self.project_path.joinpath('platformio.ini').open(mode='w') as platformio_ini_file:
                platformio_ini_config.write(platformio_ini_file)

            logger.info("'platformio.ini' has been patched")

        shutil.rmtree(self.project_path.joinpath('include'), ignore_errors=True)

        # Remove 'src' directory too but on case-sensitive file systems 'Src' == 'src' == 'SRC' so we need to check
        # first
        if not self.project_path.joinpath('SRC').is_dir():
            shutil.rmtree(self.project_path.joinpath('src'), ignore_errors=True)


    def start_editor(self, editor_command: str) -> int:
        """
        Start the editor specified by 'editor_command' with the project opened

        Args:
            editor_command: editor command as we start it in the terminal

        Returns:
            return code of the editor on success, -1 otherwise
        """

        logger.info(f"starting an editor '{editor_command}'...")

        try:
            # Works unstable on some Windows 7 systems, but correct on latest Win7 and Win10...
            # result = subprocess.run([editor_command, str(self.project_path)], check=True)
            result = subprocess.run(f"{editor_command} {str(self.project_path)}", check=True, shell=True)
            return result.returncode if result.returncode != -1 else 0
        except subprocess.CalledProcessError as e:
            logger.error(f"failed to start the editor {editor_command}: {e.stderr}")
            return -1


    def pio_build(self) -> int:
        """
        Initiate a build of the PlatformIO project by the PlatformIO ('run' command). PlatformIO prints error message
        by itself to the STDERR so there is no need to catch it and outputs by us

        Returns:
            0 if success, raise an exception otherwise
        """

        command_arr = [self.config.get('app', 'platformio_cmd'), 'run', '-d', str(self.project_path)]
        if logger.getEffectiveLevel() > logging.DEBUG:
            command_arr.append('--silent')

        result = subprocess.run(command_arr)
        if result.returncode == 0:
            logger.info("successful PlatformIO build")
        else:
            logger.error("PlatformIO build error")
        return result.returncode


    def clean(self) -> None:
        """
        Clean-up the project folder preserving only an '.ioc' file
        """

        for child in self.project_path.iterdir():
            if child.name != f"{self.project_path.name}.ioc":
                if child.is_dir():
                    shutil.rmtree(child, ignore_errors=True)
                    logger.debug(f"del {child}/")
                elif child.is_file():
                    child.unlink()
                    logger.debug(f"del {child}")

        logger.info("project has been cleaned")



"""
Some unit-tests for stm32pio. Uses sample project to generate and build it. It's OK to get errors on `test_run_editor`
one because you don't necessarily should have all of the editors. Run as

    python3 -m unittest discover -v

or

    python3 -m stm32pio.tests.test -v

(from repo's root)
"""



import setuptools

import stm32pio.app

with open("README.md", 'r') as readme:
    long_description = readme.read()

setuptools.setup(
    name='stm32pio',
    version=stm32pio.app.__version__,
    author='ussserrr',
    author_email='andrei4.2008@gmail.com',
    description="Small cross-platform Python app that can create and update PlatformIO projects from STM32CubeMX .ioc "
                "files. It uses STM32CubeMX to generate a HAL-framework based code and alongside creates PlatformIO "
                "project with the compatible framework specified",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ussserrr/stm32pio",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Topic :: Software Development :: Embedded Systems"
    ],
    python_requires='>=3.6',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'stm32pio = stm32pio.app:main'
        ]
    }
)
